(* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Routines for tabulation of the thermal rates and for the scattering term
for the Fokker-Planck type scattering operator.

Version: 12-10-2020
Contact: Andrzej Hryczuk (NCBJ)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% *)

(* Generate a table of  generic 1D function func(x) based on initial sparse grid
and then iteratively adding more points if linear and spline interpolations
differ relatively by more than "iacc" in the middle points; in log-log
Returns: table
*)
Tabulate1Dloglog[func_] := Module[
                    {add,step=0,tfunc,ilinear,icubic,lxm,tfuncadd,tableadd,tx},
  (*---------------- first tabulation ----------------- *)
  (* initial table in x -- logarithmic spacing between xmin/5 and 1000
  (extended range compared to BE solution to avoid extrapolation)
  +  more sparse up to xmax *)
  tx  =   Join[
            Table[xmin/5 ((1000/(xmin/5))^(1/(Nx-1)))^(i-1),{i,Nx}],
            Table[1000 ((xmax/(1000))^(1/(10)))^(i),{i,2,10}]
          ];
  tx=DeleteDuplicates[tx];
  (* initial tabulation *)
  tfunc  = ParallelMap[func,tx];
  table  = MapThread[List,{Log@tx,Log@tfunc}];

  (*-----------------interpolation check ---------------- *)
  While[step ==0 || Length[add]!=0 && step<imax,
   step++;add={};
   ilinear = Interpolation[table, InterpolationOrder -> 1];
   icubic  = Interpolation[table, InterpolationOrder -> 3];

   Do[
    (* define middle point: *)
    lxm = 1/2(table[[i,1]] + table[[i-1,1]]);
    (* check if the difference between order 1 and order 3 interpolations
    in the middle point is small enough  *)
    If[!MemberQ[table[[All,1]],lxm] && Abs[icubic[lxm]-ilinear[lxm]]>iacc*Abs[table[[i,1]]-table[[i-1,1]]],
        AppendTo[add, lxm]];
   ,{i,2,Length[table]}];

   tfuncadd  = ParallelMap[func,Exp@add];
   tableadd  = MapThread[List,{add,Log@tfuncadd}];

   table = Sort[Union[table,tableadd],#1[[1]]<#2[[1]] &];
  ];

  (*-----------------extension for x>xmax ---------------- *)
  (* add more points if for x>xmax there is still some change *)
  While[Abs[table[[-1,2]]-table[[-2,2]]]>0.01*Abs[table[[-1,2]]]
    && table[[-1,2]]>Log10[minnumber],
    AppendTo[table,{1+table[[-1,1]],Log[func[Exp[1+table[[-1,1]]]]]}]
  ];

 (* to avoid 1/0 and Log[0.] problems *)
 table=Select[table,#[[2]] \[Element] Reals &]/.{0.->minnumber,Indeterminate->minnumber};
];



(******************************************************************************)
(* Tabulation and interpolation of the thermally averaged annihilation
cross-section <sv> *)
PrepANN := Module[{itsv,ltsv},

    (* perform tabulation *)
    If[RelThermalAv,
     Tabulate1Dloglog[TAsvx],
     Tabulate1Dloglog[TAsvxnonrel]
    ];

    (* store the result as tsv *)
    lltsv = N[table];
    tsv = Exp@lltsv;

    (* interpolation *)
    itsv=Interpolation[lltsv,InterpolationOrder->3];

    (* create ouptut fixing the extrapolation to a constant *)
    svx[x_/;(tsv[[1,1]] <= x <= tsv[[-1,1]]) ]    := Exp[itsv[Log[x]]];
    svx[x_/;( x < tsv[[1,1]]) ]                   := Exp[itsv[lltsv[[1,1]]]];
    svx[x_/;(tsv[[-1,1]] < x) ]                   := Exp[itsv[lltsv[[-1,1]]]];

    (* save also the definition of the derivative *)
    dsvx[x_] := Derivative[1][svx][x];
];

(******************************************************************************)
(* Tabulation and interpolation of the thermally averaged p^2 weighted
annihilation cross-section <sv>_2 *)
PrepANN2 := Module[{itsv2,ltsv2},

    (* perform tabulation *)
    If[RelThermalAv,
     Tabulate1Dloglog[TAsv2x],
     Tabulate1Dloglog[TAsv2xnonrel]
    ];

    (* store the result as tsv2 *)
    lltsv2 = N[table];
    tsv2 = Exp@lltsv2;

    (* interpolation *)
    itsv2=Interpolation[lltsv2,InterpolationOrder->3];

    (* create ouptut fixing the extrapolation to a constant *)
    sv2x[x_/;(tsv2[[1,1]] <= x <= tsv2[[-1,1]]) ]   := Exp[itsv2[Log[x]]];
    sv2x[x_/;( x < tsv2[[1,1]]) ]                   := Exp[itsv2[lltsv2[[1,1]]]];
    sv2x[x_/;(tsv2[[-1,1]] < x) ]                   := Exp[itsv2[lltsv2[[-1,1]]]];

    (* save also the definition of the derivative *)
    dsv2x[x_] := Derivative[1][sv2x][x];
];

(******************************************************************************)
(* Tabulation and interpolation of the scattering rate *)
PrepSCATT := Module[{igamma,ltgamma,lltgamma},

If[!scatttype=="gamma(x)" && !FullCel,
    (* this integration is fast; for better accuracy increase Nx *)
    If[Nx<100,Nx=100];

    (* perform tabulation *)
    Tabulate1Dloglog[gamma];

    (* store the result as tgamma *)
    lltgamma = N[table];
    tgamma = Exp@lltgamma;
    lltgamma=Select[lltgamma,NumberQ[#[[2]]]&];

    (* interpolation *)
    igamma=Interpolation[lltgamma,InterpolationOrder->1];

    (* create ouptut fixing the extrapolation to a constant *)
    gam[x_/;(tgamma[[1,1]] <= x <= tgamma[[-1,1]]) ]  := Exp[igamma[Log[x]]];
    gam[x_/;( x < tgamma[[1,1]]) ]                    := Exp[igamma[lltgamma[[1,1]]]];
    gam[x_/;(tgamma[[-1,1]] < x) ]                    := minnumber;
];

];


(******************************************************************************)
(* tabulation and interpolation of the angular average of the sv *)
PrepareANNtheta := Module[{add,tp,power,pmin,pmax,pfomin,pfomax,s,step=0,
                          ltp1,ltp2,in,integr,ct0,svMol,mid,data,datafo,ltsvtheta,sv0},

  (* min and max momenta for tabulation *)
  pmin = Max[mDM/xmax*(xmin)^(1/2)*qmin/2,mDM/10^6]; (* note: very small p can lead to difficulties in the integration, while not relevant for evolution -- limit to v >= 10^-6 *)
  pmax = mDM/xmin*(xmax)^(1/2)*qmax;

  (* find maximal velocity (Vmin) that below which sv changes by less than 1 per mille *)
  GetVmin = Module[{f, vmin = 10^-10., n = 1},
   f[v_] := sv[4mDM^2(1+v^2/2)^2];

   (* first check if finite in v->0 limit *)
   sv0 = Check[NLimit[f[v], v -> 0],"sv divergent in v->0 limit"] // Quiet;

   (* only if sv->const != 0 in v->0 limit look for largest Vmin that still gives ~sv[v=0] *)
   If[NumericQ[sv0] && sv0 != 0. ,
     While[Abs[1 - sv0/f[vmin 1.2^n]] < 10^-3., n++; Vmin = vmin 1.2^(n - 1)],
     Vmin = 0.;
    ];
   ];


   (* -------------------------------------------------------------------------*)
  If[threshold=={},

    (* define the integrand and integration function *)
    integr[p1_,p2_,ct_] = Piecewise[
        {{sv0,p1<mDM Vmin && p2< mDM Vmin},
        {(1 - p1 p2 ct/(Sqrt[p1^2 + mDM^2] Sqrt[p2^2 + mDM^2])) sv[(Sqrt[p1^2 + mDM^2] + Sqrt[p2^2 + mDM^2])^2 - (p1^2 + p2^2 + 2 p1 p2 ct)],p1>=mDM Vmin || p2>= mDM Vmin}}
      ];
    in[p1_?NumericQ, p2_?NumericQ] := in[p1, p2] = in[p2, p1] =
                    NIntegrate[1/2 integr[p1, p2, ct], {ct,-1,1},
                     MinRecursion -> 1,
                     MaxRecursion -> 20,
                     WorkingPrecision -> MachinePrecision,
                     Method -> {"GlobalAdaptive", "SymbolicProcessing" -> False}
                    ];

    (* split table generation in two parts: dense table around the typical
    freeze-out values and a general more sparse one *)
    data := data =
      Reap[DensityPlot[Log[in[Exp[lp1], Exp[lp2]]],
      {lp1, Log[pmin], Log[pmax]}, {lp2,Log[pmin], Log[pmax]},
      Mesh -> All,
      PlotRange -> All,
      MaxRecursion -> plotMaxR,
      PlotPoints -> plotPoints,
      EvaluationMonitor :> Sow[{lp1, lp2, Log[in[Exp[lp1], Exp[lp2]]]}]]
      ];

    pfomin = 0.2 Sqrt[2 mDM mDM/20.];
    pfomax = 10. Sqrt[2 mDM mDM/20.];
    datafo := datafo =
      Reap[DensityPlot[Log[in[Exp[lp1], Exp[lp2]]],
      {lp1, Log[pfomin],Log[pfomax]}, {lp2, Log[pfomin], Log[pfomax]},
      Mesh -> All,
      PlotRange -> All,
      MaxRecursion -> Min[plotMaxR, 4],  (* note: here even if set smaller in settings we insist to have plotMaxR >= 4 *)
      PlotPoints -> Min[plotPoints, 40], (* note: here even if set smaller in settings we insist to have plotPoints >=40 *)
      EvaluationMonitor :> Sow[{lp1, lp2, Log[in[Exp[lp1], Exp[lp2]]]}]]
      ];

    (* perform the taublation in parallel *)
    Parallelize[{{$KernelID, data}, {$KernelID, datafo}}];

    (* combine to get the final result *)
    ltsvtheta=Join[data[[2,1]],datafo[[2,1]]];
    isvthetatmp = Check[Interpolation[(ltsvtheta)/.Indeterminate->Log10[minnumber],InterpolationOrder->1],"err"];

    (* it may happen that the mesh is too low quality for some points to give good
    interpolation; in that case perturb slightly the values and try again *)
    If[isvthetatmp=="err",
      Print["Difficulty in interpolation. Trying to solve it..."];
      ltsvtheta = {#1, #2, #3 (1 + RandomReal[10^-6 {-1, 1}])} & @@@ ltsvtheta;
      isvthetatmp = Check[Interpolation[(ltsvtheta)/.Indeterminate->Log10[minnumber],InterpolationOrder->1],"err"];

      (* if error persists write a warning and quit *)
      If[isvthetatmp=="err",Print["Error in interpolation of tsvtheta. Needs inspection..."];Abort[];];
    ];

    (* if all good define the output interpolation function *)
    isvtheta[p1_Real,p2_Real]=Exp[isvthetatmp[Log[p1],Log[p2]]];
    (* and the tabulation *)
    tsvtheta=Exp[ltsvtheta];

  , (* if thershold !={} *)

    integr[p1_,p2_,ct_] = (1 - p1 p2 ct/(Sqrt[p1^2 + mDM^2] Sqrt[p2^2 + mDM^2]))
                      sv[Max[(Sqrt[p1^2 + mDM^2] + Sqrt[p2^2 + mDM^2])^2 - (p1^2 + p2^2 + 2 p1 p2 ct),threshold[[1]]]]; (*Max in argument needed to avoid round-offs*)

    in[p1_?NumericQ, p2_?NumericQ] :=in[p1, p2] = in[p2, p1] = Module[{ctupper,ctupperI,return,tr},
      tr=threshold[[1]];
      ctupper=(2.* Sqrt[p1^2 + mDM^2]*Sqrt[p2^2 + mDM^2]+2. mDM^2-tr)/(2.*p1*p2);
      If[ctupper<=-1,return=minnumber,
        ctupperI=Min[1,ctupper];
        return=NIntegrate[1/2 integr[p1, p2, ct], {ct,-1,ctupperI},
          MinRecursion -> 1,
          MaxRecursion -> 20,
          WorkingPrecision -> MachinePrecision,
          Method -> {"GlobalAdaptive", "SymbolicProcessing" -> False}];
      ];
      return
    ];

    (* split table generation in two parts: dense table around the typical
    freeze-out values and a general more sparse one *)
    data=Reap[DensityPlot[Log[in[Exp[lp1],Exp[lp2]]],
            {lp1,Log[pmin],Log[pmax]},{lp2,Log[pmin],Log[pmax]},
            Mesh->All,
            PlotRange->All,
            MaxRecursion->plotMaxR,
            PlotPoints->plotPoints,
            EvaluationMonitor:> Sow[{lp1,lp2,Log[in[Exp[lp1],Exp[lp2]]]}]]];

    pfomin = 0.1 Sqrt[2 mDM mDM/20.];
    pfomax = 10. Sqrt[2 mDM mDM/20.];
    datafo = Reap[DensityPlot[Log[in[Exp[lp1], Exp[lp2]]],
            {lp1, Log[pfomin], Log[pfomax]},{lp2, Log[pfomin], Log[pfomax]},
            Mesh -> All,
            PlotRange -> All,
            MaxRecursion -> Min[plotMaxR,4],
            PlotPoints -> Min[plotPoints,40],
            EvaluationMonitor :> Sow[{lp1, lp2, Log[in[Exp[lp1], Exp[lp2]]]}]]];

    (* perform the taublation *)
    ltsvtheta=Join[data[[2,1]],datafo[[2,1]]];
    isvthetatmp = Interpolation[(ltsvtheta)/.Indeterminate->Log10[minnumber],InterpolationOrder->1];

    (* define the output interpolation function *)
    isvtheta[p1_Real,p2_Real]=Exp[isvthetatmp[Log[p1],Log[p2]]];
    (* and the tabulation *)
    tsvtheta=Exp[ltsvtheta];

  ];

];

(******************************************************************************)
(*
Routine regenerating the table calculated by PrepareANNtheta to structure it
on an uniform grid suitable then for linear interpolation
*)
RegArrayGen[tab_]:=Module[{lp1min,lp1max,lp2min,lp2max,dp,tp,intsvtheta},
  (* find the min and max grid values in both directions *)
  lp1min = Min[Log[tab][[All,1]]];
  lp1max = Max[Log[tab][[All,1]]];
  lp2min = Min[Log[tab][[All,2]]];
  lp2max = Max[Log[tab][[All,2]]];
  (* check if they are the same, as they should *)
  If[lp1min==lp2min && lp1max==lp2max,lpmin=lp1min;lpmax=lp1max,Print["Error: the input table for <sv>_theta is not square in p1,p2"];Break[];];

  (* create denser grid *)
  dp = (lpmax-lpmin)/(Np-1.);
  tp = Table[lpmin+(k-1.) dp,{k,Np}];
  (* create intarpolation from supplied tab in Log[{p1,p2,svteta}] *)
  intsvtheta = Check[Interpolation[Log[tab],InterpolationOrder->1],"err"];

  (* it may happen that the mesh is too low quality for some points to give good
  interpolation; in that case perturb slightly the values and try again *)
  If[intsvtheta=="err",
    Print["Difficulty in interpolation. Trying to solve it..."];
    tab = {#1, #2, #3 (1 + RandomReal[10^-6 {-1, 1}])} & @@@ tab;
    intsvtheta = Check[Interpolation[Log[tab],InterpolationOrder->1],"err"];
    If[intsvtheta=="err",Exit[];];
  ];

  (* create much denser regular array out of it; table Log[{p1,p2,svtheta}] *)
  RegArray=Array[intsvtheta[tp[[#1]],tp[[#2]]]&,{Length[tp],Length[tp]}];
  tpregsvtheta=ToPackedArray[RegArray];
  ltpregsvtheta=Length[tpregsvtheta];

];

(* ---------------------------------------------------------------------------*)
(* A wrapper to run both with one call *)
PrepANNtheta:=(PrepareANNtheta;RegArrayGen[tsvtheta]);
